/*
 *  LaTeX2Sdoc
 *  Copyright (C) 2001 SAKURAI, Masashi (m.sakurai@cmt.phys.kyushu-u.ac.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package jp.gr.java_conf.ccs2.tool.tex2sdoc;
import java.util.*;
import jp.gr.java_conf.ccs2.tool.tex2sdoc.math.MathMLCommandTable;
import jp.gr.java_conf.ccs2.tool.tex2sdoc.math.MathMLCommandHandler;
import jp.gr.java_conf.ccs2.jdk11.lang.EasyMain;
import jp.gr.java_conf.ccs2.jdk11.util.StringUtil;


public class TeXUtil extends EasyMain {

    public static void main(String [] args) {
	while(true) {
	    String test = read("Input tex source ([.]:quit)","");
	    if (test == null || test.equals("")) continue;
	    if (test.equals(".")) break;
	    print( prepareTeXSource(test) );
	}
    }

    static private final int STATE_NO = 0;
    static private final int STATE_ID = 1;
    static private final int STATE_CMD = 2;
    static private final int STATE_ENV = 3;
    /**
       normalize TeX source with separated identifier.
       Ex.  "abcd" => "a b c d"
     */
    public static String prepareTeXSource(String input) {
	//replace XML escape character
	input = StringUtil.replace(input,"&amp;","&");
	input = StringUtil.replace(input,"&lt;","<");
	input = StringUtil.replace(input,"&gt;",">");
	input = StringUtil.replace(input,"&quot;","\"");
	input = StringUtil.replace(input,"&apos;","\'");
	input = StringUtil.replace(input,"\\!","\\SPACEFACTRIAL");
	//TeX normalize
	StringBuffer out = new StringBuffer();
	int state = STATE_NO;
	int environment = 0;
	char prv = ' ';
	boolean prvLit = false;
	StringBuffer curCmd = null;
	for (int i=0;i<input.length();i++) {
	    char c = input.charAt(i);
	    boolean lit = checkLiteral(c);
	    //System.out.println(">"+c+"  "+state);
	    switch (state) {
	    case STATE_NO:
		//line separator
		if (c == '\\' && prv == '\\') {
		    out.append(c).append(whiteSpace(environment));
		    c = ' ';
		    break;
		}
		//enter command
		if (lit && prv == '\\') {
		    out.append(c);
		    curCmd = new StringBuffer();
		    curCmd.append(c);
		    state = STATE_CMD;
		    break;
		}
		//enter identifier
		if (lit && prvLit) {
		    out.append(whiteSpace(environment)).append(c);
		    state = STATE_ID;
		    break;
		}
		//upper and lower (Maybe buggy?)
		if ( (prv == '^' || prv == '_') && 
		     (c != '\\' && c != '{')) {
		    out.append(c).append(whiteSpace(environment));
		    break;
		}
		//other
		out.append(c);
		break;
	    case STATE_ID:
		//continue identifier
		if (lit) {
		    out.append(whiteSpace(environment)).append(c);
		    break;
		}
		//end of identifier zone
		out.append(c);
		state = STATE_NO;
		break;
	    case STATE_CMD:
		//continue command zone
		out.append(c);
		//end of command zone
		if (!lit) {
		    String cmdName = curCmd.toString();
		    //Text command or environment ?
		    MathMLCommandHandler hn = 
			MathMLCommandTable.getHandler(cmdName);
		    if (cmdName.equals("begin") ) {
			environment++;
		    } else  if (cmdName.equals("end") ) {
			environment--;
			break;
		    } else  if ( (hn != null) && hn.isTextMode()) {
			ArgumentInfo agi = parseMatchedBrace(input,i,1);
			i = agi.endPos-1;
			out.append(agi.contents[0]);
			out.append("}");
		    }
		    state = STATE_NO;
		}
		curCmd.append(c);
		break;
	    default:
		// NOP
	    }
	    prv = c;
	    prvLit = lit;
	}
	return out.toString();
    }

    private static String whiteSpace(int env) {
	if (env > 0) return "";
	return " ";
    }

    /** 
       make a string 2D array from an array contents of latex expression.
     */
    public static String [][] makeArray(String input) {
	String [] lines = splitLine(input);
	String [][] rets = new String[lines.length][];
	for (int i=0;i<lines.length;i++) {
	    rets[i] = splitColm(lines[i]);
	    if (rets[0].length != rets[i].length) {
		throw new RuntimeException("Bad array format.");
	    }
	}
	return rets;
    }

    /** split by "&",
       make a column array from a line contents of latex expression.
     */
    public static String [] splitColm(String line) {
	Vector colms = new Vector();
	char prv = ' ';
	int lastPos = 0;
	for (int i=0;i<line.length();i++) {
	    char c = line.charAt(i);
	    if (c == '&' && prv != '\\') {
		colms.addElement( line.substring(lastPos,i) );
		lastPos = i + 1;
	    }
	    prv = c;
	}
	colms.addElement( line.substring(lastPos) );
	String [] rets = new String[colms.size()];
	for (int i=0;i<rets.length;i++) {
	    rets[i] = (String)colms.elementAt(i);
	}
	return  rets;
    }

    /** split by "\\",
       make a line array from a array contents of latex expression.
     */
    public static String [] splitLine(String input) {
	Vector lines = new Vector();
	char prv = ' ';
	int lastPos = 0;
	for (int i=0;i<input.length();i++) {
	    char c = input.charAt(i);
	    if (c == '\\' && prv == '\\') {
		lines.addElement( input.substring(lastPos,i-1) );
		lastPos = i + 1;
		prv = ' ';
		continue;
	    }
	    prv = c;
	}
	lines.addElement( input.substring(lastPos) );
	String [] rets = new String[lines.size()];
	for (int i=0;i<rets.length;i++) {
	    rets[i] = (String)lines.elementAt(i);
	}
	return  rets;
    }

    /**
       @param text input tex source
       @param com command name with back slash (Ex. \cite)
       @param sx start position to find command 
       @return -1 or first position of the command string
     */
    public static int findCmd(String text,String com,int sx) {
	int ps = text.indexOf(com,sx);
	if (ps == -1 ) return -1;
	if (text.length() > ps + com.length()){
	    if (checkLiteral( text.charAt(ps+com.length())) ) {
		return findCmd(text,com,ps+com.length());
	    }
	}
	return ps;
    }

    /** if true, the given character is the litteral. */
    public static boolean checkLiteral(char c) {
	if ( ('\u0041' <= c && c <= '\u005a') || // Large Case 
	     ('\u0061' <= c && c <= '\u007a') || // Small Case
	     ('\u00c0' <= c && c <= '\u00d6') || 
	     ('\u00d8' <= c && c <= '\u00f6') ||
	     ('\u00f8' <= c && c <= '\u00ff') ||
	     ('\u0100' <= c && c <= '\u1fff') ||
	     ('\u3040' <= c && c <= '\u318f') ||
	     ('\u3300' <= c && c <= '\u337f') ||
	     ('\u3400' <= c && c <= '\u3d2d') ||
	     ('\u4e00' <= c && c <= '\u9fff') ||
	     ('\uf900' <= c && c <= '\ufaff') ) {
	    return true;
	}
	return false;
    }

    public static boolean checkNumber(char c) {
	if ( '0' <= c && c <= '9' ) {
	    return true;
	}
	return false;
    }

    /** 
       @param text input tex source
       @param sx start position to find the brace
       (sx don't have to be a position of the fist brace.
       This routine will find the first brace from the position sx.)
       @param args number of arguments
       @return null or ArgumentInfo object. 
       Parsed arguments don't have argument-braces.
       So don't mind chopping braces.
     */
    public static ArgumentInfo parseMatchedBrace(String text,int sx,int args) {
	return parseMatchedBrace(text,'{','}',sx,args);
    }
    
    /** 
       @param text input tex source
       @param sx start position to find the brace
       (sx don't have to be a position of the fist brace.
       This routine will find the first brace from the position sx.)
       @param leftBrace 
       @param rightBrace
       @param args number of arguments
       @return null or ArgumentInfo object. 
       Parsed arguments don't have argument-braces.
       So don't mind chopping braces.

         [ArgumentInfo] Ex.
         \text{abcd} after
               ^    ^ [endPos]
         [startPos] 
         [contents] = {abcd}
     */
    public static ArgumentInfo parseMatchedBrace(
	String text,char leftBrace,
	char rightBrace,int sx,int args) {

	if (args <= 0) return null;
	String [] ret = new String[args];
	int as = -1,ae = -1;//start and end position of whole arguments
	for( int i=0;i<args;i++) {
	    int leftPs = searchBrace(text,leftBrace,sx)+1;
	    if (leftPs == 0) return null;
	    if (as == -1) as = leftPs;
	    char prev = ' ';
	    int braces = 0;
	    for (int j=leftPs+1;j<text.length();j++) {
		char c = text.charAt(j);
		if (c == rightBrace && prev != '\\' ) {
		    if (braces == 0) {
			ret[i] = text.substring(leftPs,j);
			sx = j;
			ae = j+1;
			break;
		    }
		    braces--;
		} else if (c == leftBrace && prev != '\\') {
		    braces++;
		}
		prev = c;
	    }
	}
	return new ArgumentInfo(ret,as,ae);
    }

    public static int searchBrace(String text,String brace,int ps) {
	return searchBrace(text,brace.charAt(0),ps);
    }
    
    public static int searchBrace(String text,char brace,int ps) {
	int ops = 0;
	while(ops>=0) {
	    ops = text.indexOf(brace,ps);
	    if (ops == -1) return -1;
	    if (ops == 0) break;
	    if (text.charAt(ops-1) != '\\') break;
	}
	return ops;
    }

}
