/* LaTeX2Sdoc
 *  Copyright (C) 2000-2001 SAKURAI, Masashi 
 *  (m.sakurai@cmt.phys.kyushu-u.ac.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package jp.gr.java_conf.ccs2.tool.tex2sdoc;
import java.util.*;
import jp.gr.java_conf.ccs2.jdk11.util.*;
import jp.gr.java_conf.ccs2.jdk11.lang.*;

/**
   translate latex format to SmartDoc format.

   Time-stamp: <2001-03-02 21:48:38 sakurai>
 */

public class Tex2Sdoc {

    static String encoding = "Shift_JIS";
    static String lang = "ja";
    static boolean noAnd = false;
    static boolean noEqref = false;
    static boolean expandMacro = false;
    static MacroManager macros;

    public static void main(String [] args) {
	if (args.length == 0) 
	    usage();
	else {
	    ArrayList files = new ArrayList();
	    for (int i=0;i<args.length;i++) {
		if (check(args[i],"-encoding") != -1) {
		    i++;
		    encoding = args[i];
		    continue;
		}
		if (check(args[i],"-lang") != -1) {
		    i++;
		    lang = args[i];
		    continue;
		}
		if (check(args[i],"-exmacro") != -1) {
		    expandMacro = true;
		    macros = new MacroManager();
		    continue;
		}
		if (check(args[i],"-noeqref") != -1) {
		    noEqref = true;
		    continue;
		}
		if (check(args[i],"-noand") != -1) {
		    noAnd = true;
		    continue;
		}
		files.add(args[i]);
	    }
	    for (int i=0;i<files.size();i++) {
		String file = (String)files.get(i);
		trans(file);
		System.out.println("TeX file ["+file+"] was translated successfully.");
	    }
	}
    }

    static void usage() {
	System.out.println("TeX2Sdoc (0.2.0) 2001/03/02  (C)2001 Sakurai");
	System.out.println("% java jp.gr.java_conf.ccs2.tool.tex2sdoc.Tex2Sdoc [-encodeing (Shift_JIS)] [-lang (ja)] (TeX files)");
    }

    //===========================
    // main routine
    //===========================

    static StringBuffer out;//output string
    static int number = 0;//current line number
    static boolean documentMode = false;//is in document environment
    static boolean descriptionMode = false;//is in description mode
    static Iterator lines;//line iterator
    static String restString;//rest charactor
    static StringBuffer bibliographyLines = new StringBuffer();
    
    static void trans(String filename) {
	Vector file = ResourceFile.readFile(filename);//jp.gr.java_conf.ccs2.jdk11.util.ResourceFile
	if (file == null) {
	    System.err.println("I/O error. Couldn't translate the file, ["+filename+"]");
	    return;
	}
	out = new StringBuffer("<?xml version='1.0' encoding=\""+encoding+"\" ?>\n");
	out.append("<doc xml:lang=\""+lang+"\">\n<head>\n");
	lines = file.iterator();
	number = 0;
	while(lines.hasNext()) {
	    int ps = 0;
	    String line;
	    if (restString == null) {
		line = getNextLine();
	    } else {
		line = restString;
		restString = null;
	    }
	    if (nativeCommand(line)) continue;
	    if (title(line)) continue;
	    if (author(line)) continue;
	    if (date(line)) continue;
	    if (abstracts(line)) continue;
	    if (!documentMode) continue;
	    if (maketitle(line)) continue;
	    if (sectionCommand(line)) continue;
	    if (environment(line)) continue;
	    if (comment(line)) continue;
	    if (newpage(line)) continue;
	    if (ignoreCommand(line)) continue;
	    if (documentMode)
		out.append(escape(line)).append("\n");
	}
	//flush section tag
	while(!sectionStack.empty()) {
	    SectionPack sp = (SectionPack)sectionStack.pop();
	    out.append(sp.endTag);
	}
	out.append("</body>\n");
	out.append(bibliographyLines.toString());
	out.append("</doc>\n");

	//output
	int ps = filename.indexOf(".");
	String rep = "out.sdoc";
	if (ps != -1)
	    rep = filename.substring(0,ps)+".sdoc";
	EasyWriter wt = new EasyWriter(rep,encoding);//jp.gr.java_conf.ccs2.jdk11.lang.
	wt.print(out.toString());
	wt.close();
    }

    static String getNextLine() {
	number++;
	String s = (String)lines.next();
	if (expandMacro && documentMode)
	    s = macros.expandMacro(s);
	return s;
    }

    //===========================
    // error code
    //===========================
    static void syntaxError() {
	System.err.println("Syntax error in "+number);
    }

    //===========================
    // escape charactor
    //===========================
    static String escapeLite(String line) {
	line = mathChar(line);
	line = newline(line);
	line = bf(line);
	line = it(line);
	line = tt(line);
	line = em(line);
	line = large(line);
	line = small(line);
	line = note(line);
	line = cite(line);
	line = ref(line);
	return line;
    }

    static String escape(String line) {
	line = escapeLite(line);
	String c = StringUtil.replace(line,"&","&amp;");//jp.gr.java_conf.ccs2.jdk11.util.
	c = StringUtil.replace(c,"<","&lt;");
	c = StringUtil.replace(c,">","&gt;");
	c = StringUtil.replace(c,"\"","&quot;");
	c = StringUtil.replace(c,"\'","&apos;");
	c = StringUtil.replace(c,"`","&#xFF5E;");
	c = StringUtil.replace(c,"\","&#x2015;");
	//c = StringUtil.replace(c,"~", "&#x007E;");
	c = StringUtil.replace(c,"~", "");
	c = StringUtil.replace(c,"","&#xFFE5;");
	c = StringUtil.replace(c,"P","&#xFFE3;");
	c = StringUtil.replace(c,"_","&#xFF3C;");
	c = StringUtil.replace(c,"a","&#x2225;");
	c = StringUtil.replace(c,"|","&#xFF0D;");
	// INTIKI!!
	c = StringUtil.replace(c,ilt,"<");
	c = StringUtil.replace(c,igt,">");
	c = StringUtil.replace(c,iquote,"\"");
	return c;
    }
    static String ilt = "#INTIKI-LT";
    static String igt = "#INTIKI-GT";
    static String iquote = "#INTIKI-QUOTE";

    //===========================
    // make title
    //===========================
    static String makeElement(String line,String texName) {
	return makeElement(line,texName,texName);
    }
    static String makeElement(String line,String texName,String tagName) {
	String tt = "\\"+texName+"{";
	int ps = line.indexOf(tt);
	if (ps == -1) return null;
	int ed = line.indexOf("}",ps);
	if (ed == -1) {
	    while(lines.hasNext()) {
		line = line+escapeLite(getNextLine());
		ed = line.indexOf("}",ps);
		if (ed >= 0) break;
	    }
	    if (ed == -1) {
		syntaxError();
		throw new InternalError();
	    }
	}
	//rest string
	if (ed < (line.length()-1))
	    restString = line.substring(ed);
	return line.substring(0,ps)+
	    "<"+tagName+">"+
	    escape(line.substring(ps+tt.length(),ed))+
	    "</"+tagName+">\n";
    }

    static boolean title(String line) {
	String tt = makeElement(line,"title");
	if (tt == null) return false;
	out.append(tt);
	return true;
    }
    static boolean author(String line) {
	String tt = makeElement(line,"author");
	if (tt == null) return false;
	out.append(tt);
	return true;
    }
    static boolean date(String line) {
	String tt = makeElement(line,"date");
	if (tt == null) return false;
	out.append(tt);
	return true;
    }
    static boolean abstracts(String line) {
	String tt = makeElement(line,"abstract");
	if (tt == null) return false;
	out.append(tt);
	return true;
    }
    static boolean maketitle(String line) {
	String tt = "\\maketitle";
	int ps = line.indexOf(tt);
	if (ps == -1) return false;
	out.append("</head>\n<body>\n");
	//rest string
	int ed = ps+tt.length();
	if (ed < (line.length()-1))
	    restString = line.substring(ed);
	return true;
    }

    //===========================
    // section 
    //===========================
    static boolean sectionCommand(String line) {
	if (section(line)) return true;
	if (subsection(line)) return true;
	if (part(line)) return true;
	if (chapter(line)) return true;
	if (subsubsection(line)) return true;
	if (appendix(line)) return true;
	return false;
    }

    static Stack sectionStack = new Stack();
    static String section(String line,String tex,
			  String tag,String attr,int lev) {
 	String tt = "\\"+tex+"{";
	int ps = line.indexOf(tt);
	if (ps == -1) {
	    tt = "\\"+tex+"*{";
	    ps = line.indexOf(tt);
	    if (ps == -1) return null;
	}
	int ed = line.indexOf("}",ps);
	if (ps == -1) {
	    syntaxError();
	    throw new InternalError();
	}
	String end = "</"+tag+">\n\n";
	while(!sectionStack.empty()) {
	    SectionPack sp = (SectionPack)sectionStack.peek();
	    if (lev <= sp.level) {
		sectionStack.pop();
		out.append(sp.endTag);
	    } else
		break;
	}
	sectionStack.push(new SectionPack(end,lev));
	//rest string
	//if (ed < (line.length()-1))
	//    restString = line.substring(ed);
	return line.substring(0,ps)+
	    "<"+tag+" "+attr+"=\""+escape(line.substring(ps+tt.length(),ed))+
	    "\""+label(line)+">\n";
    }

    static boolean part(String line) {
	String r = section(line,"part","part","title",1);
	if (r == null) return false;
	out.append(r);
	return true;
    }
    static boolean chapter(String line) {
	String r = section(line,"chapter","chapter","title",2);
	if (r == null) return false;
	out.append(r);
	return true;
    }
    static boolean section(String line) {
	String r = section(line,"section","section","title",3);
	if (r == null) return false;
	out.append(r);
	return true;
    }
    static boolean subsection(String line) {
	String r = section(line,"subsection","subsection","title",4);
	if (r == null) return false;
	out.append(r);
	return true;
    }
    static boolean subsubsection(String line) {
	String r = section(line,"subsubsection","subsubsection","title",5);
	if (r == null) return false;
	out.append(r);
	return true;
    }
    static boolean appendix(String line) {
	String r = section(line,"appendix","appendix","title",0);
	if (r == null) return false;
	out.append(r);
	return true;
    }

    //===========================
    // begin and end [environment]
    //===========================

    static boolean environment(String line) {
	if (equation(line)) return true;
	if (item(line)) return true;
	if (quote(line)) return true;
	if (figure(line)) return true;
	if (itemize(line)) return true;
	if (enumerate(line)) return true;
	if (description(line)) return true;
	if (table(line)) return true;
	if (tabbing(line)) return true;
	if (centerAlign(line)) return true;
	if (leftAlign(line)) return true;
	if (rightAlign(line)) return true;
	if (bibliography(line)) return true;
	return false;
    }
    /**
       check the beginning of environment and make a sdoc tag as an answer.
       @param line line string
       @param tex environment name
       @param tag sdoc tag name
     */
    static String begin(String line,String tex,String tag) {
	return begin(line,tex,tag,null);
    }

    /**
       check the beginning of environment and make a sdoc tag as an answer.
       @param line line string
       @param tex environment name (Ex. itemize)
       @param tag sdoc tag name (Ex. dl, ul, ol,)
       @param attr sdoc attribute (Ex. style="text-align:right")
     */
    static String begin(String line,String tex,String tag,String attr) {
	String tt = "\\begin{"+tex+"}";
	int ps = line.indexOf(tt);
	if (ps == -1) return null;
	//rest string
	int ed = ps+tt.length();
	if (ed < (line.length()-1))
	    restString = line.substring(ed);
	if (StringUtil.isNull( tag ))
	    return escape(line.substring(0,ps))+"<"+tag+">\n";
	return escape(line.substring(0,ps))+"<"+tag+" "+attr+">\n";
    }

    /**
       check the end of environment and make a sdoc tag as an answer.
       @param line line string
       @param tex environment name
       @param tag sdoc tag name
     */
    static String end(String line,String tex,String tag) {
	String tt = "\\end{"+tex+"}";
	int ps = line.indexOf(tt);
	if (ps == -1) return null;
	//rest string
	int ed = ps+tt.length();
	if (ed < (line.length()-1))
	    restString = line.substring(ed);
	return escape(line.substring(0,ps))+"</"+tag+">\n\n";
    }

    static boolean centerAlign(String line) {
	String r = end(line,"center","div");
	if (r != null)
	    out.append(r);
	String p = begin(line,"center","div","style=\"text-align:center\"");
	if (p!=null)
	    out.append(p);
	return (r!=null)||(p!=null);
    }

    static boolean rightAlign(String line) {
	String r = end(line,"flushright","div");
	if (r != null)
	    out.append(r);
	String p = begin(line,"flushright","div",
			 "style=\"text-align:right\"");
	if (p!=null)
	    out.append(p);
	return (r!=null)||(p!=null);
    }

    static boolean leftAlign(String line) {
	String r = end(line,"flushleft","div");
	if (r != null)
	    out.append(r);
	String p = begin(line,"flushleft","div",
			 "style=\"text-align:left\"");
	if (p!=null)
	    out.append(p);
	return (r!=null)||(p!=null);
    }

    static boolean itemize(String line) {
	String r = end(line,"itemize","ol");
	if (r != null)
	    out.append(r);
	String p = begin(line,"itemize","ol",label(line));
	if (p!=null)
	    out.append(p);
	return (r!=null)||(p!=null);
    }

    static boolean enumerate(String line) {
	String r = end(line,"enumerate","ul");
	if (r != null)
	    out.append(r);
	String p = begin(line,"enumerate","ul",label(line));
	if (p!=null)
	    out.append(p);
	return (r!=null)||(p!=null);
    }

    static boolean description(String line) {
	String r = end(line,"description","dl");
	if (r != null)
	    out.append(r);
	String p = begin(line,"description","dl",label(line));
	if (p!=null)
	    out.append(p);
	return (r!=null)||(p!=null);
    }
    static boolean item(String line) {
	String tt = "\\item";
	int ps = line.indexOf(tt);
	if (ps == -1) return false;
	if (descriptionMode) {
	    //description
	    int dlPos = tt.length()+ps;
	    ArgumentInfo dtArg = TeXUtil.parseMatchedBrace(
		line,'[',']',dlPos,1);
	    if (dtArg != null) {
		dlPos = dtArg.endPos;
		out.append("<dt>"+dtArg.contents[0]+"</dt>\n");
	    }
	    out.append("<dd>"+line.substring(dlPos)+"</dd>\n");
	} else {
	    //itemize and enumerate
	    out.append("<li>"+
		       escape(line.substring(ps+tt.length()))+
		       "</li>\n");
	}
	return true;
    }
    static boolean quote(String line) {
	String r = end(line,"quote","blockquote");
	if (r != null)
	    out.append(r);
	String p = begin(line,"quote","blockquote");
	if (p!=null)
	    out.append(p);
	return (r!=null)||(p!=null);
    }

    //===========================
    // comment
    //===========================
    static boolean comment(String line) {
	String tt = "%";
	int ps = line.indexOf(tt);
	if (ps == -1) return false;
	if (ps >= 2) return false;
	out.append(escape(line.substring(0,ps))+
		   "<!-- "+escape(line.substring(ps+tt.length()))+" -->\n");
	return true;
    }

    //===========================
    // newpage command
    //===========================
    static boolean newpage(String line) {
	String [] tts = {"\\newpage","\\clearpage"};
	int ps = -1;
	for (int i=0;i<tts.length;i++) {
	    ps = line.indexOf(tts[i]);
	    if (ps != -1) break;
	    if (ps == -1 && i == (tts.length-1)) return false;
	}
	out.append(escape(line.substring(0,ps)));
	out.append("<or><native format=\"latex\">\\clearpage</native><native format=\"html\"><hr></native></or>\n\n");
	return true;
    }

    //===========================
    // inline decoration
    //===========================
    static String makeInlineEmptyTag(String tagname,String attr) {
	if (StringUtil.isNull(attr))
	    return ilt+tagname+"/"+igt;
	return ilt+tagname+" "+StringUtil.replace(attr,"\"",iquote)+"/"+igt;
    }
    static String makeInlineBTag(String tagname,String attr) {
	if (StringUtil.isNull(attr))
	    return ilt+tagname+igt;
	return ilt+tagname+" "+StringUtil.replace(attr,"\"",iquote)+igt;
    }
    static String makeInlineETag(String tagname) {
	return ilt+"/"+tagname+igt;
    }
    static String makeInlineTag(String tagname,String attr,String content) {
	return makeInlineBTag(tagname,attr)+content+makeInlineETag(tagname);
    }
    static String inline(String line,String begin,String end,String tag) {
	return inline(line,begin,end,tag,null);
    }
    static String inline(String line,String begin,String end,
			 String tag,String attr) {
	while(true) {
	    int ps = line.indexOf(begin);
	    if (ps == -1) return line;
	    int ed = line.indexOf(end,ps);
	    if (ed == -1) {
		//get more line
		while(lines.hasNext()) {
		    line = line+escapeLite(getNextLine());
		    ed = line.indexOf(end,ps);
		    if (ed >= 0) break;
		}
		if (ed == -1) {
		    syntaxError();
		    throw new InternalError();
		}
	    }
	    String ns = makeInlineBTag(tag,attr)+
		line.substring(ps+begin.length(),ed)+
		makeInlineETag(tag);
	    line = line.substring(0,ps)+ns+line.substring(ed+end.length());
	}
    }
    static String bf(String line) {
	line = inline(line,"{\\bfseries","}","b");
	line = inline(line,"\\textbf{","}","b");
	line = inline(line,"{\\bf","}","b");
	return inline(line,"{\\sffamily","}","b");
    }
    static String it(String line) {
	line = inline(line,"\\textit{","}","i");
	line = inline(line,"\\textsl{","}","i");
	line = inline(line,"{\\itshape","}","i");
	line = inline(line,"{\\slshape","}","i");
	return inline(line,"{\\it","}","i");
    }
    static String tt(String line) {
	line = inline(line,"\\texttt{","}","tt");
	line = inline(line,"{\\ttfamily","}","tt");
	return inline(line,"{\\tt","}","tt");
    }
    static String em(String line) {
	line = inline(line,"\\emph{","}","em");
	return inline(line,"{\\em","}","em");
    }
    static String large(String line) {
	line = inline(line,"{\\large","}","span","style=\"font-size:120%\"");
	line = inline(line,"{\\Large","}","span","style=\"font-size:144%\"");
	line = inline(line,"{\\LARGE","}","span","style=\"font-size:173%\"");
	line = inline(line,"{\\huge","}","span","style=\"font-size:208%\"");
	return inline(line,"{\\Huge","}","span","style=\"font-size:250%\"");
    }
    static String small(String line) {
	line = inline(line,"{\\small","}","span","style=\"font-size:83%\"");
	line = inline(line,"{\\footnotesize","}","span","style=\"font-size:69%\"");
	line = inline(line,"{\\scriptsize","}","span","style=\"font-size:58%\"");
	return inline(line,"{\\tiny","}","span","style=\"font-size:48%\"");
    }
    static String note(String line) {
	return inline(line,"\\footnote{","}","note");
    }
    static String cite(String line) {
	while(true) {
	    int ps = TeXUtil.findCmd(line,"\\cite",0);
	    if (ps == -1) return line;
	    ArgumentInfo cites = TeXUtil.parseMatchedBrace(line,ps,1);
	    if (cites == null) {
		syntaxError();
		throw new InternalError();
	    }
	    String [] names = StringUtil.split(cites.contents[0],",");
	    if (names == null || names.length == 0) {
		syntaxError();
		System.err.println("Bug found in cite command.");
		throw new InternalError();
	    }
	    String out = line.substring(0,ps);
	    for (int i=0;i<names.length;i++)
		out += makeInlineEmptyTag("a","href=\"#"+names[i]+"\"");
	    out += line.substring(cites.endPos);
	    line = out;
	}
    }
    static String ref(String line) {
	String [] refCmds = {"\\ref","\\eqref"};
	int ps = -1;
	while(true) {
	    for (int i=0;i<refCmds.length;i++) {
		ps = TeXUtil.findCmd(line,refCmds[i],0);
		if (ps == -1 && i == (refCmds.length-1)) return line;
		if (ps != -1) break;
	    }
	    ArgumentInfo argRef = TeXUtil.parseMatchedBrace(line,ps,1);
	    if (argRef == null) {
		syntaxError();
		throw new InternalError();
	    }
	    String out = line.substring(0,ps);
	    if (!noEqref) {
		out += makeInlineEmptyTag(
		    "a","href=\"#"+argRef.contents[0]+"\"");
	    }
	    out += line.substring(argRef.endPos);
	    line = out;
	}
    }
    static String newline(String line) {
	String tt = "\\\\";
	String rr = makeInlineEmptyTag("br",null);
	return StringUtil.replace(line,tt,rr);
    }
    //MATH CHARACTOR
    static String mathChar(String line) {
	String begin = "$";
	String end = "$";
	while(true) {
	    int ps = line.indexOf(begin);
	    if (ps == -1) return line;
	    int ed = line.indexOf(end,ps+1);
	    if (ed == -1) {
		//get more line
		while(lines.hasNext()) {
		    line = line+escapeLite(getNextLine());
		    ed = line.indexOf(end,ps+1);
		    if (ed >= 0) break;
		}
		if (ed == -1) {
		    syntaxError();
		    throw new InternalError();
		}
	    }
	    String mathContent = line.substring(ps+begin.length(),ed);
	    line = 
		line.substring(0,ps)+
		makeInlineBTag("span","normalizer=\"tex\"")+
		mathContent+makeInlineETag("span")+
		line.substring(ed+end.length());
	}
    }

    //===========================
    // table
    //===========================
    static boolean table(String line) {
	int beginNumber = number;
	String tt = "\\begin{tabular}";
	int ps = check(line,tt);
	if (ps == -1) return false;
	out.append(escape(line.substring(0,ps))+"\n");
	out.append("<table"+label(line)+">\n<tbody>\n");
	//
	boolean error = true;
    outer:
	while(lines.hasNext()) {
	    line = getNextLine();
	    while(check(line,"\\\\") == -1) {
		line += getNextLine();
		ps = check(line,"\\end{tabular}");
		if (ps != -1) {
		    parseLine(line.substring(0,ps));
		    error = false;
		    break outer;
		}
	    }
	    parseLine(line);
	}
	if (error) {
	    System.err.println("not found a terminator of Tabular environment (begin from "+beginNumber+").");
	    throw new InternalError();
	}
	//
	out.append("</tbody>\n</table>\n");
	return true;
    }
    static void parseLine(String line) {
	out.append("<tr><td>");
	int ps = -1,last=0;
	while(true) {
	    ps = line.indexOf("&",last);
	    if (ps == -1) {
		//terminal?
		ps = line.indexOf("\\\\",last);
		if (ps  == -1) {
		    out.append(escape(line.substring(last)));
		    break;
		}
		out.append(escape(line.substring(last,ps)));
		break;
	    }
	    out.append(escape(line.substring(last,ps)));
	    out.append(" </td><td> ");
	    last = ps + 1;
	}
	out.append(" </td></tr>\n");
    }

    //===========================
    // tabbing
    //===========================
    static boolean tabbing(String line) {
	int beginNumber = number;
	String tt = "\\begin{tabbing}";
	int ps = check(line,tt);
	if (ps == -1) return false;
	out.append(escape(line.substring(0,ps))+"\n");
	out.append("<table"+label(line)+">\n<tbody>\n");
	//
	boolean error = true;
    outer:
	while(lines.hasNext()) {
	    line = getNextLine();
	    while(check(line,"\\\\") == -1) {
		line += getNextLine();
		ps = check(line,"\\end{tabbing}");
		if (ps != -1) {
		    parseLine(line.substring(0,ps));
		    error = false;
		    break outer;
		}
	    }
	    parseTabline(line);
	}
	if (error) {
	    System.err.println("not found a terminator of Tabbing environment (begin from "+beginNumber+").");
	    throw new InternalError();
	}
	//
	out.append("</tbody>\n</table>\n");
	return true;
    }
    static void parseTabline(String line) {
	if (line.indexOf("\\kill") != -1) return;
	out.append("<tr><td>");
	int ps = -1,last=0;
	while(true) {
	    ps = line.indexOf("\\>",last);
	    if (ps == -1) {
		//terminal?
		ps = line.indexOf("\\\\",last);
		if (ps  == -1) {
		    out.append(escape(line.substring(last)));
		    break;
		}
		out.append(escape(line.substring(last,ps)));
		break;
	    }
	    out.append(escape(line.substring(last,ps)));
	    out.append(" </td><td> ");
	    last = ps + 1;
	}
	out.append(" </td></tr>\n");
    }

    //===========================
    // equation and eqnarray
    //===========================
    static boolean equation(String line) {
	if (equation_gen(line)) return true;
	if (eqnarray_gen(line)) return true;
	return false;
    }
    static boolean equation_gen(String line) {
	int beginNumber = number;
	String tt = "\\begin{equation}";
	int ps = check(line,tt);
	if (ps == -1) return false;
	out.append(escape(line.substring(0,ps))+"\n");
	out.append("<equation"+label(line)+">\n");
	//out.append("<![CDATA[\n");
	//
	equationContent("equation");
	//
	//out.append("]]>\n");
	out.append("</equation>\n");
	return true;
    }

    static boolean eqnarray_gen(String line) {
	String tt = "\\begin{eqnarray}";
	int ps = check(line,tt);
	if (ps == -1) return false;
	out.append(escape(line.substring(0,ps))+"\n");
	out.append("<equation>\n");
	//out.append("<![CDATA[\n");
	out.append("\\begin{array}{rcl}\n");
	//
	equationContent("eqnarray");
	//
	out.append("\\end{array}\n");
	//out.append("]]>\n");
	out.append("</equation>\n");
	return true;
    }

    static void equationContent(String envEnd) {
	int beginNumber = number;
	boolean error = true;
	while(lines.hasNext()) {
	    String line = getNextLine();
	    int ps = check(line,"\\end{"+envEnd+"}");
	    if (ps != -1) {
		out.append(noAnd(line.substring(0,ps))).append("\n");
		error = false;
		break;
	    }
	    out.append(noAnd(line)).append("\n");
	}
	if (error) {
	    System.err.println("not found a terminator of ["+envEnd+"] environment (begin from "+beginNumber+").");
	    throw new InternalError();
	} 
    }

    static String noAnd(String in) {
	if (noAnd)
	    return StringUtil.replace(in,"&"," ");
	return in;
    }

    // if found label, return id attribute. if not, return "".
    static String label(String line) {
	String label = "";
	int labelPos = TeXUtil.findCmd(line,"\\label",0);
	if (labelPos != -1) {
	    ArgumentInfo argLabel = TeXUtil.parseMatchedBrace(line,labelPos,1);
	    if (argLabel == null) {
		syntaxError();
		throw new InternalError();
	    }
	    label = " id=\""+argLabel.contents[0]+"\"";
	}
	return label;
    }

    //===========================
    // Bibliography
    //===========================
    static boolean bibliography(String line) {
	int beginNumber = number;
	String tt = "\\begin{thebibliography}";
	int ps = check(line,tt);
	if (ps == -1) return false;
	out.append(escape(line.substring(0,ps))+"\n");
	bibliographyLines.append("<bibliography>\n");
	//
	String prvLine = getNextLine();
	Stack curLines = new Stack();
	boolean error = true;
	while(lines.hasNext()) {
	    line = getNextLine();
	    int bibs = check(line,"bibitem");
	    int ites = check(line,"item");
	    if (bibs == -1 && ites == -1) {
		ps = check(line,"\\end{thebibliography}");
		//out
		if (ps != -1) {
		    error = false;
		    break;
		}
		//continue
		if (!curLines.isEmpty()) curLines.pop();
		prvLine += line;
		curLines.push(parseBibitem(prvLine));
		continue;
	    }
	    curLines.push(parseBibitem(line));
	    prvLine = line;
	}
	if (error) {
	    System.err.println("not found a terminator of Bibliography environment (begin from "+beginNumber+").");
	    throw new InternalError();
	}
	//
	while(!curLines.isEmpty()) {
	    bibliographyLines.append( (String)curLines.firstElement() );
	    curLines.removeElementAt(0);
	}
	bibliographyLines.append("</bibliography>\n");
	return true;
    }

    static String parseBibitem(String line) {
	StringBuffer ret = new StringBuffer();
	String bibitem = "\\bibitem";
	int ps = line.indexOf(bibitem);
	//bibitem
	if (ps != -1) {
	    ArgumentInfo argName = TeXUtil.parseMatchedBrace(
		line,ps+bibitem.length(),1);
	    if (argName == null) {
		syntaxError();
		throw new InternalError();
	    }
	    ret.append("<book id=\""+argName.contents[0]+"\">");
	    ret.append("<title>"+line.substring(argName.endPos)+"</title>");
	    ret.append("</book>\n");
	    return ret.toString();
	}
	//item
	bibitem = "\\item";
	ps = line.indexOf(bibitem);
	ret.append("<book>");
	ret.append("<title>"+line.substring(ps+bibitem.length())+"</title>");
	ret.append("</book>\n");
	return ret.toString();
    }

    //===========================
    // figure 
    //===========================
    static boolean figure(String line) {
	int beginNumber = number;
	String tt = "\\begin{figure}";
	int ps = check(line,tt);
	if (ps == -1) return false;
	out.append(escape(line.substring(0,ps))+"\n");
	//
	String caption = "";
	String label = "";
	String option = "";
	String file = "";
	boolean error = true;
	while(lines.hasNext()) {
	    line = getNextLine();
	    //includegraphics
	    ps = TeXUtil.findCmd(line,"\\includegraphics",0);
	    if (ps != -1) {
		ArgumentInfo argOpt = 
		    TeXUtil.parseMatchedBrace(line,'[',']',ps,1);
		if (argOpt != null) {
		    String [] opts = StringUtil.split(argOpt.contents[0],",");
		    if (opts != null && opts.length != 0) {
			for (int i=0;i<opts.length;i++) {
			    option += " "+StringUtil.replace(
				opts[i],"=","=\"")+"\"";
			}
		    }
		}
		ArgumentInfo argFile = 
		    TeXUtil.parseMatchedBrace(line,ps,1);
		if (argFile == null) {
		    syntaxError();
		    throw new InternalError();
		}
		file = " src=\""+argFile.contents[0]+"\"";
	    }
	    //caption
	    ps = TeXUtil.findCmd(line,"\\caption",0);
	    if (ps != -1) {
		ArgumentInfo argCap = 
		    TeXUtil.parseMatchedBrace(line,ps,1);
		if (argCap == null) {
		    syntaxError();
		    throw new InternalError();
		}
		file = " title=\""+argCap.contents[0]+"\"";
	    }
	    //label
	    ps = TeXUtil.findCmd(line,"\\label",0);
	    if (ps != -1) {
		ArgumentInfo argLbl = 
		    TeXUtil.parseMatchedBrace(line,ps,1);
		if (argLbl == null) {
		    syntaxError();
		    throw new InternalError();
		}
		file = " id=\""+argLbl.contents[0]+"\"";
	    }
	    //end
	    ps = check(line,"\\end{figure}");
	    if (ps != -1) {
		error = false;
		break;
	    }
	}
	if (error) {
	    System.err.println("not found a terminator of Figure environment (begin from "+beginNumber+").");
	    throw new InternalError();
	}
	//
	out.append("<figure "+file+label+caption+option+"/>\n");
	return true;
    }

    //===========================
    // save native command
    //===========================
    static boolean nativeCommand(String line) {
	int ps = 0;
	while(true) {
	    //set
	    ps = check(line,"\\set");
	    if (ps != -1) break;
	    // def
	    ps = check(line,"\\def");
	    if (ps != -1) {
		if (expandMacro)
		    macros.addMacro(line);
		break;
	    }
	    // newcommand
	    ps = check(line,"\\newcommand");
	    if (ps != -1) {
		if (expandMacro)
		    macros.addMacro(line);
		break;
	    }
	    //noindent
	    ps = check(line,"\\noindent");
	    if (ps != -1) break;
	    //document
	    if (check(line,"{document}") != -1) {
		documentMode = !documentMode;
		return true;
	    }
	    //skip
	    ps = check(line,"\\bigskip");
	    if (ps != -1) break;
	    ps = check(line,"\\medskip");
	    if (ps != -1) break;
	    ps = check(line,"\\smallskip");
	    if (ps != -1) break;
	    //not found
	    return false;
	} 
	out.append(escape(line.substring(0,ps)));
	out.append("<native format=\"latex\">"+
		   line.substring(ps)+"</native>\n");
	return true;
    }
    
    //===========================
    // ignore command
    //===========================
    static boolean ignoreCommand(String line) {
	//unknown environment
	String tt = "\\begin{";
	int ps = check(line,tt);
	if (ps != -1) {
	    int ed = line.indexOf("}",ps);
	    System.err.println("Unknown environment ["+line.substring(ps+tt.length(),ed)+"]");
	    return true;
	}
	if (check(line,"\\end{") != -1) return true;
	return false;
    }

    static int check(String line,String word) {
	return line.indexOf(word);
    }

}

class SectionPack {
    String endTag;
    int level;
    SectionPack(String e,int l) {
	endTag = e; level = l;
    }
}
