/* LaTeX2Sdoc
 *  Copyright (C) 2000-2001 SAKURAI, Masashi (m.sakurai@cmt.phys.kyushu-u.ac.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package jp.gr.java_conf.ccs2.tool.tex2sdoc.math;
import java.util.*;

public class MathMLTransTable {

    public static String transGreek(String tex, boolean ent) {
	String code = "&"+getResourceString(tex,GREEK)+";";
	if (ent) return code;
	return transEntity(code);
    }
    public static String transCharacter(String tex, boolean ent) {
	String code = getResourceString(tex,CHARACTER);
	if (ent) return code;
	return transEntity(code);
    }
    public static String transFunction(String tex, boolean ent) {
	String code = getResourceString(tex,FUNCTION);
	if (ent) return code;
	return transEntity(code);
    }
    
    public static String transEntity( String code) {
	if (code == null) return null;
	int sp = code.indexOf("&");
	int ep = code.indexOf(";");
	if ( sp != -1 && ep != -1) {
	    code = code.substring(sp+1,ep);
	    return getResourceString(code,ENTITY);
	}
	return code;
    }

    //==(load resource file)=====

    private static String [] names = {"Greek","Character","Function","Entity"};
    private static int GREEK = 0;
    private static int CHARACTER = 1;
    private static int FUNCTION = 2;
    private static int ENTITY = 3;

    private static ResourceBundle [] resources 
               = new ResourceBundle[names.length];
    private static boolean [] loadFails = new boolean[names.length];

    /** loading a transfer resource file */
    private static void loadResources(int i) {
	try {
	    resources[i] = ResourceBundle.getBundle(
		"jp.gr.java_conf.ccs2.tool.tex2sdoc.math.MathML"+names[i]);
	} catch (MissingResourceException mre) {
	    System.err.println("[MathML"+names[i]+".properties] not found");
	    loadFails[i] = true;
	}
    }

    /** get resource string from properties file "MathMLSymbol.properties" */
    private  static String getResourceString(String s,int i) {
	if (resources[i] == null && !loadFails[i]) 
	    loadResources(i);
	try {
	    return resources[i].getString(s);
	} catch (RuntimeException e) {
	    //just return null!
	}
	return null;
    }

}
