/* LaTeX2Sdoc
 *  Copyright (C) 2000-2001 SAKURAI, Masashi (m.sakurai@cmt.phys.kyushu-u.ac.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package jp.gr.java_conf.ccs2.tool.tex2sdoc.math;
import java.util.*;
import jp.gr.java_conf.ccs2.jdk11.util.StringUtil;

/** Navigate TeXSourceVisitor object. (Not thread safe.) */
public class TeXSourceVisitorCenter {
    
    protected TeXSourceVisitor visitor;
    protected Expression root;

    /** just construct object */
    public TeXSourceVisitorCenter() {
    }

    /** construct and traverse */
    public TeXSourceVisitorCenter(Expression object,
				  TeXSourceVisitor visitor) throws ParseException {
	traverse(object,visitor);
    }

    /** traverse object */
    public void traverse(Expression object,TeXSourceVisitor visitor) throws ParseException {
	this.root = object;
	this.visitor = visitor;
	this.visitor.init();
	visitExpression( this.root );
    }

    void visitExpression( Expression exp ) throws ParseException {
	if (exp == null) return;
	Term [] terms = exp.getTerms();
	if (terms == null || terms.length == 0) 
	    return; //maybe error?

	visitor.expressionBegin();

	for (int i=0;i<terms.length;i++) {
	    visitTerm( terms[i] );
	}
	
	visitor.expressionEnd();
    }

    void visitTerm( Term term ) throws ParseException {
	Element body = term.getBodyElement();
	Element sub = term.getSubElement();
	Element sup = term.getSupElement();
	
	boolean decoration = true;
	if (sub == null && sup == null) {
	    decoration = false;
	} else {
	    if (sub != null && sup != null) {
		visitor.expressionSubsupBegin();
	    } else {
		if (sub != null) {
		    visitor.expressionSubBegin();
		} else if (sup != null) {
		    visitor.expressionSupBegin();
		}
	    }
	}

	//body
	visitElement( body );
   
	//decoration
	if (sub != null)
	    visitElement( sub );
	if (sup != null)
	    visitElement( sup );
	if (decoration)
	    visitor.expressionEnd();
    }

    void visitElement( Element element ) throws ParseException {
	if (element instanceof Expression) {
	    visitExpression( (Expression)element );
	} else if (element instanceof Content) {
	    Content content = (Content)element;
	    if (content instanceof NumberContent) {
		visitor.number( content.getContent() );
	    } else if (content instanceof IdentifierContent) {
		visitor.identifier( content.getContent() );
	    } else if (content instanceof OperatorContent) {
		visitor.operator( content.getContent() );
	    } else if (content instanceof CommandContent) {
		visitCommand( (CommandContent)content );
	    } else if (content instanceof EnvironmentContent) {
		EnvironmentContent env = (EnvironmentContent)content;
		visitor.environment( env );
	    } else {
		throw new InternalError(
		    "TeX2MathML bug. No such subclass of Content. ["+content+"] in "+this.getClass().getName());//error
	    }
	} else {
	    throw new InternalError("TeX2MathML big. No such subclass of Element. ["+element+"] in "+this.getClass().getName());//error
	}
    }

    void visitCommand( CommandContent cmd ) throws ParseException {
	visitor.commandBegin( cmd );

	if ( !StringUtil.isNull(cmd.getOption()) )
	    visitor.commandOption();

	Element [] args = cmd.getArguments();
	if (args != null && args.length != 0) {
	    for (int i=0;i<args.length;i++) {
		visitor.argBegin();
		visitElement( args[i] );
		visitor.argEnd();
	    }
	}
	visitor.commandEnd();
    }

}
