/* LaTeX2Sdoc
 *  Copyright (C) 2000-2001 SAKURAI, Masashi (m.sakurai@cmt.phys.kyushu-u.ac.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package jp.gr.java_conf.ccs2.tool.tex2sdoc.math.handler;
import jp.gr.java_conf.ccs2.tool.tex2sdoc.math.*;
import jp.gr.java_conf.ccs2.tool.tex2sdoc.*;
import jp.gr.java_conf.ccs2.jdk11.util.StringUtil;

/**
   TeX -> MathML Array Environment handler
 */
public class ArrayEnvHandler implements MathMLEnvHandler {

    String content;
    String [] align;

    public ArrayEnvHandler(String content) {
	ArgumentInfo info = TeXUtil.parseMatchedBrace(content,0,1);
	this.content = content.substring(info.endPos);

	String a = info.contents[0];//alignment
	String [] as = StringUtil.split(TeXUtil.prepareTeXSource(a));
	align = new String[as.length];
	for (int i=0;i<as.length;i++) {
	    if (as[i].indexOf("l") != -1) {
		align[i] = "left";
	    } else if (as[i].indexOf("r") != -1) {
		align[i] = "right";
	    } else {
		align[i] = "center";
	    }
	}
    }

    public void onEnvBegin(MathMLVisitor parent) {
	parent.addString( TagUtil.makeBtag(parent,"mtable") );
    }

    public void onEnvEnd(MathMLVisitor parent) {
	parent.addString( TagUtil.makeEtag(parent,"mtable") );
    }

    public void onContentOut(MathMLVisitor parent) throws ParseException {
	String [][] array = TeXUtil.makeArray(content);
	TeX2MathML eng = new TeX2MathML();
	//
	for(int i=0;i<array.length;i++) {
	    parent.addString( TagUtil.makeBtag(parent,"mtr") );
	    parent.incTab();
	    for (int j=0;j<array[0].length;j++) {
		parent.addString( TagUtil.makeBtagWithAttr(
		    parent,"mtd","columnalign",align[j]) );
		String out = eng.translate_gen(array[i][j],parent);
		parent.addString( out );
		parent.addString( TagUtil.makeEtag(parent,"mtd") );
	    }
	    parent.decTab();
	    parent.addString( TagUtil.makeEtag(parent,"mtr") );
	}
	parent.log(eng.getLog());
    }

}
